unit SUtil;

interface

uses Windows;

{ Type Convert Function }

function IStr(Value: Integer): string;
function FStr(const Value: Double; Digits: Integer): string;
function IZStr(Value, Width: Integer): string;
function SInt(const S: string): Integer;
function SFloat(const S: string): Double;
function UnitFormat(Value: Comp): string;
function NumFormat(Value: Integer): string;

{ Strings Function }

function Lower(const S: string): string;
function Upper(const S: string): string;
function LStr(const S: string; Count: Integer): string;
function RStr(const S: string; Count: Integer): string;
function RevStr(const S: string): string;
function StrColl(const S1, S2: string): Integer;
function ConvLineBreak(const S, LineBreak: string): string;
function CopyCStr(const Src: string; Dst: PChar; Size: Cardinal): Cardinal;

{ Path Function }

function IsFile(const Path: string): Boolean;
function IsDir(const Path: string): Boolean;
function AbsPath(const Path: string): string;
procedure SplitDrive(const Path: string; var Drive, Tail: string);
procedure SplitPath(const Path: string; var Head, Tail: string);
function BaseName(const Path: string): string;
function DirName(const Path: string): string;
procedure SplitExt(const Path: string; var Root, Ext: string);
function JoinPath(const Path1, Path2: string): string;
function ExpandVars(const Path: string): string;

{ Operating System Services }

function GetEnv(const Name: string; const Default: string): string;
function SysError(Code: Integer): string;

{ Misc. }

function OsVersionInfo: TOsVersionInfo;
function LocaleStr(lcid: LCID; LCType: LCTYPE): string;
function DateStr(const Date: TSystemTime): string;
function TimeStr(const Time: TSystemTime): string;

function LongFileName(const FileName: string): string;
function GetFindData(const Path: string;
  var FindData: TWin32FindData): Boolean;

{ Classes }

type
  TListDir = class(TObject)
  private
    FHandle: THandle;
    FResult: Boolean;
    FFindData: TWin32FindData;
    FIsDir: Boolean;
  public
    constructor Create(const Path: string);
    constructor CreateWithMask(const Path: string; Mask: string);
    destructor Destroy; override;
    function Each(var FileName: string): Boolean;
    property IsDir: Boolean read FIsDir;
  end;

  TSFile = class(TObject)
  private
    FHandle: THandle;
    FOpened: Boolean;
    FName: string;
    FBuf: PChar;
    ReadP, EndP: PChar;
    function GetSize: Integer;
  public
    constructor Create(const FileName: string; const Mode: string);
    destructor Destroy; override;
    function Read(var Buffer; Count: Integer): Integer;
    function ReadLine(var S: string): Boolean;
    function Write(var Buffer; Count: Integer): Integer;
    function Seek(Offset: Integer): Integer;
    property Size: Integer read GetSize;
    property Opened: Boolean read FOpened;
    property Name: string read FName;
  end;

const
  MaxGetEnvLen = 1024; // Ō̃k܂
  MaxExpandVarsLen = 1024; // Ō̃k܂

var
  LeadBytes: set of Char = [];


implementation

procedure InitLeadBytes;
var
  cpinfo: TCPInfo;
  i, j: Integer;
begin
  GetCPInfo(CP_ACP, cpinfo);
  i := 0;
  with cpinfo do
    while ((LeadByte[i] or LeadByte[i+1]) <> 0) and (i<MAX_LEADBYTES) do
    begin
      for j := LeadByte[i] to LeadByte[i+1] do
        Include(LeadBytes, Char(j));
      Inc(i, 2);
    end;
end;


{ Type Convert Function }

function IStr(Value: Integer): string;
begin
  Str(Value, Result);
end;

function FStr(const Value: Double; Digits: Integer): string;
begin
  Str(Value:0:Digits, Result);
end;

function IZStr(Value, Width: Integer): string;
var
  sign: Boolean;
  p: PChar;
begin
  sign := Value < 0;
  if sign then
  begin
    Value := -Value;
    Dec(Width);
  end;
  Str(Value:Width, Result);
  p := PChar(Result);
  while p^ = ' ' do
  begin
    p^ := '0';
    Inc(p);
  end;
  if sign then
    Result := '-' + Result;
end;

function SInt(const S: string): Integer;
var
  code: Integer;
begin
  Val(S, Result, code);
end;

function SFloat(const S: string): Double;
var
  code: Integer;
begin
  Val(S, Result, code);
end;

// 4(_܂)+Pʕɂ
function UnitFormat(Value: Comp): string;
const
  KiroSize = 1024;
  MegaSize = 1024 * 1024;
  GigaSize = 1024 * 1024 * 1024;
var
  u: Char;
  f: Integer;
  n: Double;
begin
  if Value < KiroSize then
    Str(Value:0:0, Result)
  else begin
    if Value >= MegaSize * 1000 then
    begin
      n := Value / GigaSize;
      u := 'G';
    end
    else if Value >= KiroSize * 1000 then
    begin
      n := Value / MegaSize;
      u := 'M';
    end
    else begin
      n := Value / KiroSize;
      u := 'K';
    end;

    // ľܓStrŐ؂̂Ă邽
    if n >= 100 then
    begin
      f := 0;
      n := n - 0.5;
    end
    else if n >= 10 then
    begin
      f := 1;
      n := Trunc(n * 10) / 10;
    end
    else begin
      f := 2;
      n := Trunc(n * 100) / 100;
    end;
    Str(n:0:f, Result);
    Result := Result + u;
  end;
end;

function NumFormat(Value: Integer): string;
var
  nf: TNumberFmt;
  buf: array[0..255] of Char;
begin
  nf.NumDigits := 0;
  nf.LeadingZero := 0;
  nf.Grouping := 3;
  nf.lpDecimalSep := '.';
  nf.lpThousandSep := ',';
  nf.NegativeOrder := 1;
  GetNumberFormat(0, 0, PChar(IStr(Value)), @nf, buf, SizeOf(buf));
  Result := buf;
end;


{ String Function }

function Lower(const S: string): string;
begin
  SetString(Result, PChar(S), Length(S));
  CharLowerBuff(PChar(Result), Length(Result));
end;

function Upper(const S: string): string;
begin
  SetString(Result, PChar(S), Length(S));
  CharUpperBuff(PChar(Result), Length(Result));
end;

function LStr(const S: string; Count: Integer): string;
var
  i: Integer;
begin
  if Count >= Length(S) then
  begin
    Result := S;
    Exit;
  end;
  i := 0;
  while i < Count do
  begin
    Inc(i);
    if S[i] in LeadBytes then
      Inc(i);
  end;
  if i > Count then
    Dec(i, 2);
  Result := Copy(S, 1, i);
end;

function RStr(const S: string; Count: Integer): string;
var
  i: Integer;
begin
  if Count >= Length(S) then
  begin
    Result := S;
    Exit;
  end;
  Count := Length(S) - Count;
  i := 0;
  while i < Count do
  begin
    Inc(i);
    if S[i] in LeadBytes then
      Inc(i);
  end;
  Result := Copy(S, i+1, Length(S)-i);
end;

function RevStr(const S: string): string;
var
  i, len: Integer;
begin
  len := Length(S);
  SetString(Result, nil, len);
  i := 0;
  while i < len do
  begin
    Inc(i);
    if S[i] in LeadBytes then
    begin
      Result[len-i] := S[i];
      Result[len-i+1] := S[i+1];
      Inc(i);
    end else
      Result[len-i+1] := S[i];
  end;
end;

function StrColl(const S1, S2: string): Integer;
begin
  Result := lstrcmp(PChar(S1), PChar(S2));
end;

function ConvLineBreak(const S: string; const LineBreak: string): string;
var
  sp, ep, p: PChar;
  lblen, n: Integer;
begin
  lblen := Length(LineBreak);
  SetString(Result, nil, Length(S)+lblen-1);
  sp := @S[1];
  ep := @S[Length(S)+1];
  p := sp;
  n := 1;
  while p < ep do
  begin
    if p^ in [#10, #13] then
    begin
      Move(sp^, Result[n], p-sp);
      Inc(n, p-sp);
      Move(LineBreak[1], Result[n], lblen);
      Inc(n, lblen);
      if (p^ = #13) and ((p+1)^ = #10) then
        Inc(p);
      Inc(p);
      SetLength(Result, Length(Result)+lblen-1);
      sp := p;
    end else
      Inc(p);
  end;
  Move(sp^, Result[n], p-sp);
  Inc(n, p-sp);
  SetLength(Result, n-1);
end;

function CopyCStr(const Src: string; Dst: PChar; Size: Cardinal): Cardinal;
begin
  Result := Length(Src);
  if Result > (Size-1) then
    Result := Size-1;
  Move(PChar(Src)^, Dst^, Result);
  Dst[Result] := #0;
  Inc(Result);
end;

{ Path Function }

const
  PathSep = '\';
  ExtSep = '.';

function IsFile(const Path: string): Boolean;
var
  r: DWORD;
begin
  r := GetFileAttributes(PChar(Path));
  Result := (r <> $FFFFFFFF) and ((r and FILE_ATTRIBUTE_DIRECTORY) = 0);
end;

function IsDir(const Path: string): Boolean;
var
  r: DWORD;
begin
  r := GetFileAttributes(PChar(Path));
  Result := (r <> $FFFFFFFF) and ((r and FILE_ATTRIBUTE_DIRECTORY) <> 0);
end;

function AbsPath(const Path: string): string;
var
  buf: array [0..MAX_PATH] of Char;
  p: PChar;
  len: Integer;
begin
  len := GetFullPathName(PChar(Path), SizeOf(buf), buf, p);
  if (RStr(buf, 1) = PathSep) and (len > 3) then
    Dec(len);
  SetString(Result, buf, len);
end;

procedure SplitDrive(const Path: string; var Drive, Tail: string);
begin
  if (Length(Path) > 1) and (Path[2] = ':') then
  begin
    Drive := Copy(Path, 1, 2);
    Tail := Copy(Path, 3, Length(Path)-2);
  end else
  begin
    Drive := '';
    Tail := Path;
  end;
end;

procedure SplitPath(const Path: string; var Head, Tail: string);
var
  i, dpos: Integer;
  p: string;
begin
  SplitDrive(Path, Head, p);
  dpos := 0;
  i := 0;
  while i < Length(p) do
  begin
    Inc(i);
    if p[i] in LeadBytes then
      Inc(i)
    else if p[i] = PathSep then
      dpos := i;
  end;
  Head := Head + Copy(p, 1, dpos-1);
  if (dpos = 1) or (dpos = 2) then  // C:\hoge or \\hoge
    Head := Head + PathSep;
  Tail := Copy(p, dpos+1, Length(p)-dpos);
end;

function BaseName(const Path: string): string;
var
  s: string;
begin
  SplitPath(Path, s, Result);
end;

function DirName(const Path: string): string;
var
  s: string;
begin
  SplitPath(Path, Result, s);
end;

procedure SplitExt(const Path: string; var Root, Ext: string);
var
  i: Integer;
  head, tail: string;
begin
  SplitPath(Path, head, tail);
  for i := Length(tail) downto 1 do
    if tail[i] = ExtSep then
    begin
      if (head <> '') and (RStr(head, 1) <> PathSep) then
        head := head + PathSep;
      Root := head + Copy(tail, 1, i-1);
      Ext := Copy(tail, i, Length(tail)-i+1);
      Exit;
    end;
  Root := Path;
  Ext := '';
end;

function JoinPath(const Path1: string; const Path2: string): string;
begin
  if RStr(Path1, 1) = PathSep then
    Result := Path1 + Path2
  else
    Result := Path1 + PathSep + Path2;
end;

function ExpandVars(const Path: string): string;
var
  buf: array[0..MaxExpandVarsLen-1] of Char;
begin
  // ExpandEnvironmentStringsWin2kƕA9xƃoCgԂ̂
  // Ԃl𗘗płȂ:-(
  ExpandEnvironmentStrings(PChar(Path), buf, SizeOf(buf));
  Result := buf;
end;


{ Operation System Services }

function GetEnv(const Name: string; const Default: string): string;
var
  buf: array[0..MaxGetEnvLen-1] of Char;
begin
  SetString(Result, Buf, GetEnvironmentVariable(PChar(Name), Buf,
    SizeOf(Buf)));
  if Result = '' then
    Result := Default;
end;

function SysError(Code: Integer): string;
var
  buf: PChar;
begin
  SetString(Result, buf, FormatMessage(
    FORMAT_MESSAGE_FROM_SYSTEM or FORMAT_MESSAGE_ALLOCATE_BUFFER or $FF,
    nil, Code, (SUBLANG_DEFAULT shl 10) or LANG_NEUTRAL, PChar(@buf), 0, nil));
  LocalFree(HLOCAL(buf));
  if Result = '' then
    Result := 'Unknown error: ' + IStr(Code);
end;


{ Misc. }

function OsVersionInfo: TOsVersionInfo;
begin
  Result.dwOSVersionInfoSize := SizeOf(Result);
  GetVersionEx(Result);
end;

function LocaleStr(lcid: LCID; LCType: LCTYPE): string;
begin
  // -1:null
  SetString(Result, nil, GetLocaleInfo(lcid, LCType, nil, 0)-1);
  GetLocaleInfo(lcid, LCType, PChar(Result), Length(Result));
end;

function DateStr(const Date: TSystemTime): string;
begin
  // -1:null
  SetString(Result, nil, GetDateFormat(LOCALE_USER_DEFAULT, DATE_LONGDATE,
    @Date, nil, nil, 0)-1);
  GetDateFormat(LOCALE_USER_DEFAULT, DATE_LONGDATE, @Date, nil, PChar(Result),
                Length(Result));
end;

function TimeStr(const Time: TSystemTime): string;
begin
  SetString(Result, nil, GetTimeFormat(LOCALE_USER_DEFAULT, 0, @Time,
    nil, nil, 0)-1); // -1:null
  GetTimeFormat(LOCALE_USER_DEFAULT, 0, @Time, nil, PChar(Result),
    Length(Result));
end;

function LongFileName(const FileName: string): string;
var
  h: THandle;
  finddata: TWin32FindData;
  path: string;
begin
  Result := '';
  if FileName = '' then
    Exit;
  path := AbsPath(FileName);
  while True do
  begin
    h := FindFirstFile(PChar(path), finddata);
    if h = INVALID_HANDLE_VALUE then
    begin
      Result := path + Result;
      Exit;
    end;
    FindClose(h);

    // '.' 狤L(@Win95a)??
    if string(finddata.cFileName) = '.' then
    begin
      Result := path + Result;
      Exit;
    end;

    path := DirName(path);
    if (path = '\\') or ((Length(path) = 3) and (path[2] = ':')) then
    begin
      Result := path + finddata.cFileName + Result;
      Exit;
    end;

    Result := '\' + string(finddata.cFileName) + Result;
  end;
end;

function GetFindData(const Path: string;
  var FindData: TWin32FindData): Boolean;
var
  h: THandle;
begin
  FillChar(FindData, SizeOf(FindData), 0);
  h := FindFirstFile(PChar(Path), FindData);
  Result := h <> INVALID_HANDLE_VALUE;
  if Result then
    FindClose(h);
end;


{ TListDir }

constructor TListDir.Create(const Path: string);
begin
  CreateWithMask(Path, '*.*');
end;

constructor TListDir.CreateWithMask(const Path: string; Mask: string);
begin
  FHandle := FindFirstFile(PChar(Path+'\'+Mask), FFindData);
  FResult := FHandle <> INVALID_HANDLE_VALUE;
end;

destructor TListDir.Destroy;
begin
  if FHandle <> INVALID_HANDLE_VALUE then
    FindClose(FHandle);
  inherited;
end;

function TListDir.Each(var FileName: string): Boolean;
begin
  Result := FResult;
  if FResult then
  begin
    FileName := FFindData.cFileName;
    FIsDir := (FFindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) <> 0;
    FResult := FindNextFile(FHandle, FFindData);
  end;
end;


{ TSFile }

constructor TSFile.Create(const FileName: string; const Mode: string);
var
  access, sharemode, creation: DWORD;
begin
  sharemode := FILE_SHARE_READ or FILE_SHARE_WRITE;
  if Mode = 'w' then
  begin
    access := GENERIC_WRITE;
    creation := CREATE_ALWAYS;
    if FileName = '' then
    begin
      FName := '<STDOUT>';
      FHandle := GetStdHandle(STD_OUTPUT_HANDLE);
    end;
  end else
  begin
    access := GENERIC_READ;
    creation := OPEN_EXISTING;
    if FileName = '' then
    begin
      FName := '<STDIN>';
      FHandle := GetStdHandle(STD_INPUT_HANDLE);
    end;
  end;
  if FileName <> '' then
  begin
    FName := FileName;
    FHandle := CreateFile(PChar(FName), access, sharemode, nil, creation,
                          FILE_ATTRIBUTE_NORMAL, 0);
  end;
  FOpened := FHandle <> INVALID_HANDLE_VALUE;
end;

destructor TSFile.Destroy;
begin
  if FOpened then
    CloseHandle(FHandle);
  if FBuf <> nil then
    FreeMem(FBuf);
end;

function TSFile.GetSize: Integer;
begin
  Result := GetFileSize(FHandle, nil);
end;

function TSFile.Read(var Buffer; Count: Integer): Integer;
begin
  if not ReadFile(FHandle, Buffer, Count, DWORD(Result), nil) then
    Result := -1;
end;

function TSFile.Write(var Buffer; Count: Integer): Integer;
begin
  if not WriteFile(FHandle, Buffer, Count, DWORD(Result), nil) then
    Result := -1;
end;

function TSFile.ReadLine(var S: string): Boolean;
const
  bufsize = 16384;
var
  len: Integer;
  sp: PChar;
  foundlf: Boolean;
begin
  if FBuf = nil then;
    GetMem(FBuf, bufsize+1); // +1: for sentinel

  S := '';
  foundlf := False;
  while not foundlf do
  begin
    if ReadP = EndP then
    begin
      len := Read(FBuf^, bufsize);
      if len < 1 then
      begin
        Result := Length(S) <> 0;
        Exit;
      end;
      ReadP := FBuf;
      EndP := FBuf + len;
    end;

    sp := ReadP;
    while (ReadP < EndP) and (not foundlf) do
    begin
      foundlf := ReadP^ = #10;
      Inc(ReadP);
    end;

    S := S + Copy(sp, 1, ReadP - sp);
  end;
  Result := True;
end;

function TSFile.Seek(Offset: Integer): Integer;
begin
  Result := SetFilePointer(FHandle, Offset, nil, FILE_BEGIN);
end;

initialization
  InitLeadBytes;

end.
