unit ImgScaler;

interface

function CalcPitch(Width: Integer; Bpp: Integer): Integer;
procedure Nearest(SrcBits, DestBits: Pointer; SrcWidth, SrcHeight: Integer;
  Bpp: Integer; DestWidth, DestHeight: Integer;
  ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer);
procedure Bilinear24(SrcBits, DestBits: Pointer; SrcWidth, SrcHeight: Integer;
  DestWidth, DestHeight: Integer;
  ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer);
procedure Average24(SrcBits, DestBits: Pointer; SrcWidth, SrcHeight: Integer;
  DestWidth, DestHeight: Integer;
  ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer);

implementation

type
  PByte = ^Byte;
  PWord = ^Word;
  PTriByte = ^TTriByte;
  TTriByte = array[0..2] of Byte;
  PInteger = ^Integer;
  PByteArray = ^TByteArray;
  TByteArray = array[0..MaxInt-1] of Byte;
  PIntArray = ^TIntArray;
  TIntArray = array[0..(MaxInt div SizeOf(Integer))-1] of Integer;
  PPointerArray = ^TPointerArray;
  TPointerArray = array[0..MaxInt div SizeOf(Pointer)-1] of Pointer;

function CalcPitch(Width: Integer; Bpp: Integer): Integer;
begin
  Result := ((Width * Bpp + 7) shr 3 + 3) and not 3;
end;

procedure Nearest(SrcBits, DestBits: Pointer; SrcWidth, SrcHeight: Integer;
  Bpp: Integer; DestWidth, DestHeight: Integer;
  ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer);
var
  srcpitch, viewpitch: Integer;
  zx, zy: Double;
  sx: PIntArray;
  sy: PPointerArray;
  sp: PByteArray;
  dline, dp: PByte;
  x, y: Integer;
begin
  srcpitch := CalcPitch(SrcWidth, Bpp);
  viewpitch := CalcPitch(ViewWidth, Bpp);

  // 摜̑ΉsNZ߂
  zx := SrcWidth / DestWidth;
  zy := SrcHeight / DestHeight;
  GetMem(sx, SizeOf(Integer) * ViewWidth);
  GetMem(sy, SizeOf(Pointer) * ViewHeight);
  for x := 0 to ViewWidth-1 do
    sx[x] := Trunc(zx * (ViewLeft+x+0.5));
  for y := 0 to ViewHeight-1 do
    sy[y] := @PByteArray(SrcBits)[Trunc(zy * (ViewTop+y+0.5)) * srcpitch];

  dline := DestBits;
  case Bpp of
    1:
      begin
        FillChar(DestBits^, viewpitch * ViewHeight, 0);
        for y := 0 to ViewHeight-1 do
        begin
          sp := sy[y];
          x := 0;
          while x < ViewWidth do
          begin
            PByteArray(dline)[x shr 3] := PByteArray(dline)[x shr 3] or
               (sp[sx[x] shr 3] shl (sx[x] and 7) and $80) shr (x and 7);
            Inc(x);
          end;
          Inc(dline, viewpitch);
        end;
      end;
    4:
      for y := 0 to ViewHeight-1 do
      begin
        dp := dline;
        sp := sy[y];
        x := 0;
        while x < ViewWidth do
        begin
          dp^ := (sp[sx[x] shr 1] shl ((sx[x] and 1) * 4) and $F0) or
            (sp[sx[x+1] shr 1] shl ((sx[x+1] and 1) * 4) and $F0 shr 4);
          Inc(dp);
          Inc(x, 2);
        end;
        Inc(dline, viewpitch);
      end;
    8:
      for y := 0 to ViewHeight-1 do
      begin
        dp := dline;
        sp := sy[y];
        x := 0;
        while x < ViewWidth do
        begin
          dp^ := sp[sx[x]];
          Inc(dp);
          Inc(x);
        end;
        Inc(dline, viewpitch);
      end;
    16:
      for y := 0 to ViewHeight-1 do
      begin
        dp := dline;
        sp := sy[y];
        x := 0;
        while x < ViewWidth do
        begin
          PWord(dp)^ := PWord(@sp[sx[x]*2])^;
          Inc(dp, 2);
          Inc(x);
        end;
        Inc(dline, viewpitch);
      end;
    24:
      for y := 0 to ViewHeight-1 do
      begin
        dp := dline;
        sp := sy[y];
        x := 0;
        while x < ViewWidth do
        begin
          PTriByte(dp)^ := PTriByte(@sp[sx[x]*3])^;
          Inc(dp, 3);
          Inc(x);
        end;
        Inc(dline, viewpitch);
      end;
    32:
      for y := 0 to ViewHeight-1 do
      begin
        dp := dline;
        sp := sy[y];
        x := 0;
        while x < ViewWidth do
        begin
          PInteger(dp)^ := PInteger(@sp[sx[x]*4])^;
          Inc(dp, 4);
          Inc(x);
        end;
        Inc(dline, viewpitch);
      end;
  end;

  FreeMem(sx);
  FreeMem(sy);
end;

procedure Bilinear24(SrcBits, DestBits: Pointer; SrcWidth, SrcHeight: Integer;
  DestWidth, DestHeight: Integer;
  ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer);
var
  srcpitch, viewpitch: Integer;
  zx, zy, t, tx, ty: Double;
  sx, sy, fx2: PIntArray;
  fx, fy: PByteArray;
  fytmp: Byte;
  x, y, nextline: Integer;
  dline, dp: PByte;
  sp, sp0, sp1, sp2, sp3: PByteArray;
  f0, f1, f2, f3: Integer;
  esx, esy, tt: Integer;
begin
  // 24bpp only
  srcpitch := CalcPitch(SrcWidth, 24);
  viewpitch := CalcPitch(ViewWidth, 24);

  // 摜̑ΉsNZ߂
  zx := SrcWidth / DestWidth;
  zy := SrcHeight / DestHeight;
  // ␳l
  tx := 0.5-(1/zx)/2.0;
  ty := 0.5-(1/zy)/2.0;

  GetMem(sx, SizeOf(Integer) * ViewWidth);
  GetMem(fx, SizeOf(Byte) * ViewWidth);
  GetMem(fx2, SizeOf(Integer) * ViewWidth);
  for x := 0 to ViewWidth-1 do
  begin
    t := zx * (ViewLeft+x+tx);
    if t < 0 then
      t := 0;
    tt := Trunc(t * $100);
    sx[x] := (tt shr 8) * 3;
    fx[x]  := tt and $FF;
    fx2[x]  := $100 - fx[x];
  end;

  GetMem(sy, SizeOf(Integer) * ViewHeight);
  GetMem(fy, SizeOf(Byte) * ViewHeight);
  for y := 0 to ViewHeight-1 do
  begin
    t := zy * (ViewTop+y+ty);
    if t < 0 then
      t := 0;
    tt := Trunc(t * $100);
    sy[y] := (tt shr 8) * srcpitch;
    fy[y]  := tt and $FF;
  end;

  esx := (SrcWidth-1) * 3;
  esy := (SrcHeight-1) * srcpitch;
  dline := DestBits;
  for y := 0 to ViewHeight-1 do
  begin
    dp := dline;
    sp := PByteArray(@PByteArray(SrcBits)[sy[y]]);
    if sy[y] < esy then
      nextline := srcpitch
    else
      nextline := 0;
    //  +---+---+
    //  |sp2|sp3|  fy
    //  +---+---+  ^
    //  |sp0|sp1|  |
    //  +---+---+  +->fx
    // sp0*(1-fx)*(1-fy) + sp1*fx*(1-fy) + sp2*(1-fx)*fy + sp3*fx*fy;
    fytmp := fy[y];
    x := 0;
    while x < ViewWidth do // forwhilêقc
    begin
      f0 := fx2[x] * ($100-fytmp);
      f1 := fx[x]  * ($100-fytmp);
      f2 := fx2[x] * fytmp;
      f3 := fx[x]  * fytmp;

      sp0 := PByteArray(@sp[sx[x]]);
      if sx[x] < esx then
        sp1 := PByteArray(@sp0[3])
      else
        sp1 := sp0;
      sp2 := PByteArray(@sp0[nextline]);
      sp3 := PByteArray(@sp1[nextline]);

      dp^ := (sp0[0]*f0 + sp1[0]*f1 + sp2[0]*f2 + sp3[0]*f3) shr 16;
      Inc(dp);
      dp^ := (sp0[1]*f0 + sp1[1]*f1 + sp2[1]*f2 + sp3[1]*f3) shr 16;
      Inc(dp);
      dp^ := (sp0[2]*f0 + sp1[2]*f1 + sp2[2]*f2 + sp3[2]*f3) shr 16;
      Inc(dp);

      Inc(x);
    end;
    Inc(dline, ViewPitch);
  end;

  FreeMem(sx);
  FreeMem(sy);
  FreeMem(fx);
  FreeMem(fx2);
  FreeMem(fy);
end;

procedure Average24(SrcBits, DestBits: Pointer; SrcWidth, SrcHeight: Integer;
  DestWidth, DestHeight: Integer;
  ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer);
var
  srcPitch, viewpitch: Integer;
  zx, zy: Double;
  sx, sy, fx, fy: PIntArray;
  x, y, yy, nx, ny: Integer;
  px, py, px2, py2: Word;
  sp0, dline, dp: PByte;
  ssp, sp: PByteArray;
  i, j, ssx, ssy: Integer;
  dr, dg, db, fb ,fg, fr: Integer;
  yb, ybuf: PIntArray;
  lastline: Boolean;
begin
  // 24bpp only
  srcpitch := CalcPitch(SrcWidth, 24);
  viewpitch := CalcPitch(ViewWidth, 24);

  // 摜̑ΉsNZ߂
  zx := SrcWidth / DestWidth;
  zy := SrcHeight / DestHeight;
  // ݂̃sNZƎ̃sNZ̍KvȂ̂Ŏۂ̃TCY+1܂ŕKv
  GetMem(sx, SizeOf(Integer) * ViewWidth + 1);
  GetMem(fx, SizeOf(Integer) * ViewWidth + 1);
  for x := 0 to ViewWidth do
  begin
    sx[x] := Trunc((zx*(ViewLeft+x)) * $10000) shr 16;
    fx[x] := Trunc((zx*(ViewLeft+x)) * $10000) and $FFFF;
  end;
  GetMem(sy, SizeOf(Integer) * ViewHeight + 1);
  GetMem(fy, SizeOf(Integer) * ViewHeight + 1);
  for y := 0 to ViewHeight do
  begin
    sy[y] := Trunc((zy*(ViewTop+y)) * $10000) shr 16;
    fy[y] := Trunc((zy*(ViewTop+y)) * $10000) and $FFFF;
  end;

  nx := Trunc($10000 / zx);
  ny := Trunc($10000 / zy);

  // xɕςsNZ̃obt@
  GetMem(ybuf, (Trunc(zy)+2)*SizeOf(Integer)*3);

  dline := DestBits;
  for y := 0 to ViewHeight-1 do
  begin
    dp := dline;
    sp := PByteArray(@PByteArray(SrcBits)[sy[y]*SrcPitch]);
    yy := sy[y+1]-sy[y];
    py := $10000 - fy[y];
    py2 := fy[y+1];
    ssy := yy-Ord(py<>0)-1;
    lastline := ViewTop+y = DestHeight-1;

    x := 0;
    while x < ViewWidth do
    begin
      yb := ybuf;
      px := $10000 - fx[x];
      px2 := fx[x+1];
      ssx := sx[x+1]-sx[x]-1-Ord(px<>0);
      for i := 0 to yy do
      begin
        if lastline then
          ssp := PByteArray(@sp[SrcPitch*(i-1)])
        else
          ssp := PByteArray(@sp[SrcPitch*i]);
        sp0 := @ssp[sx[x]*3];

        dr := 0;
        dg := 0;
        db := 0;
        fr := 0;
        fg := 0;
        fb := 0;

        if px <> 0 then
        begin
          Inc(fb, sp0^ * px);
          Inc(sp0);
          Inc(fg, sp0^ * px);
          Inc(sp0);
          Inc(fr, sp0^ * px);
          Inc(sp0);
        end;
        for j := 0 to ssx do
        begin
          Inc(db, sp0^);
          Inc(sp0);
          Inc(dg, sp0^);
          Inc(sp0);
          Inc(dr, sp0^);
          Inc(sp0);
        end;
        if px2 <> 0 then
        begin
          Inc(fb, sp0^ * px2);
          Inc(sp0);
          Inc(fg, sp0^ * px2);
          Inc(sp0);
          Inc(fr, sp0^ * px2);
        end;

        // ؂̂ĂƏkƂɌ덷傫Ȃ̂
        // $8000Ďľܓ
        yb[0] := ((db + fb shr 16) * nx + $8000) shr 16;
        yb[1] := ((dg + fg shr 16) * nx + $8000) shr 16;
        yb[2] := ((dr + fr shr 16) * nx + $8000) shr 16;
        yb := PIntArray(@yb[3]);
      end;

      dr := 0;
      dg := 0;
      db := 0;
      fr := 0;
      fg := 0;
      fb := 0;

      yb := ybuf;
      if py <> 0 then
      begin
        Inc(fb, yb[0] * py);
        Inc(fg, yb[1] * py);
        Inc(fr, yb[2] * py);
        yb := PIntArray(@yb[3]);
      end;
      for j := 0 to ssy do
      begin
        Inc(db, yb[0]);
        Inc(dg, yb[1]);
        Inc(dr, yb[2]);
        yb := PIntArray(@yb[3]);
      end;
      if py2 <> 0 then
      begin
        Inc(fb, yb[0] * py2);
        Inc(fg, yb[1] * py2);
        Inc(fr, yb[2] * py2);
      end;

      // ؂̂ĂƏkƂɌ덷傫Ȃ̂
      // $8000Ďľܓ
      dp^ := ((db + fb shr 16) * ny + $8000) shr 16;
      Inc(dp);
      dp^ := ((dg + fg shr 16) * ny + $8000) shr 16;
      Inc(dp);
      dp^ := ((dr + fr shr 16) * ny + $8000) shr 16;
      Inc(dp);

      Inc(x);
    end;
    Inc(dline, viewpitch);
  end;

  FreeMem(sx);
  FreeMem(sy);
  FreeMem(fx);
  FreeMem(fy);
  FreeMem(ybuf);
end;

end.
