program NkV;

uses
  Windows, Messages, ShellAPI,
  SUtil, SCont, SForm, SCmd, SConfFile, SProf,
  Command, NkVImage, ImgFilter, SpiFilter;

{$R *.RES}
{$R nkvres.res}
{$I nkv.inc}

const
  AppName = 'NkV';
  Version = '0.5.2 (2004-11-23)';
  Copyright = 'Copyright (C) 1999-2004 Nobuyuki Honda (Ferdia)';
  License = 'This software is provided under the zlib/libpng license.';

type
  PFileItem = ^TFileItem;
  TFileItem = record
    Name: string;
    Loadable: Boolean;
  end;

  TFileList = class(TObject)
  private
    FItems: TSList;
    CurPos: Integer;
    function Get(Index: Integer): PFileItem;
    function GetCount: Integer;
  public
    Parent: TFileList;
    constructor Create;
    destructor Destroy; override;
    procedure Add(Item: PFileItem);
    procedure Clear;
    procedure Update(const Dir: string; const Ar: TArchive);
    function IndexOf(const FileName: string): Integer;
    property Items[Index: Integer]: PFileItem read Get; default;
    property Count: Integer read GetCount;
  end;

  TNkVMainForm = class(TSForm)
    NormalCursor: HCURSOR;
    GrabCursor: HCURSOR;
    OldPt: TPoint;
    ZoomValue: Double;
    ViewLeft, ViewTop, DispWidth, DispHeight: Integer;

    Stretch: Boolean;
    ScaleFilter: Boolean;
    ShowGrid: Boolean;
    ShowColorValue: Boolean;

    FileName: string;
    FileCTime: TFileTime;
    FileATime: TFileTime;
    FileMTime: TFileTime;

    procedure InitParams; override;
    procedure CalcViewSize(var Width: Integer; var Height: Integer);
    procedure Zoom(Value: Double);
    procedure Scroll(X, Y: Integer);
    function Load(const AFileName: string; const AAr: TArchive; Quiet: Boolean): Boolean;
    function LoadNext(const AFileList: TFileList; const Ar: TArchive; Direction: Integer): Boolean;
    function LoadFile(const Path: string): Boolean;
    function LoadNextFile(Direction: Integer; LeaveAr: Boolean): Boolean;
    procedure UpdateCaption;
    procedure UpdateView;
    procedure FitWindow;
    procedure DoCmd(const CmdName: string);
    procedure ShowAbout;
    procedure PopupMenu(const Pt: TPoint);
    procedure Error(const AFileName: string; const Msg: string);
    procedure SaveAs;
    procedure SetWallPaper(Tile: Boolean);
    procedure CopyFileNameToClipBoard;
    procedure ShowPluginConfig(const PluginFileName: string);
    procedure ShowAboutPlugins;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure WMSetCursor(var Message: TWMSetCursor); message WM_SETCURSOR;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
    procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
    procedure WMMouseWheel(var Message: TMessage); message WM_MOUSEWHEEL;
    procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMDropFiles(var Message: TWMDropFiles); message WM_DROPFILES;
    procedure WMCommand(var Message: TWMCommand); message WM_COMMAND;
    procedure WMContextMenu(var Message: TMessage); message WM_CONTEXTMENU;
  end;

  TPicInfoDlg = class(TSModelessDlg)
    procedure SetItems;
    procedure WMInitDialog(var Message: TWMInitDialog); message WM_INITDIALOG;
    procedure WMCommand(var Message: TWMCommand); message WM_COMMAND;
    procedure WMDrawItem(var Message: TWMDrawItem); message WM_DRAWITEM;
  end;

const
  IDM_PLUGINSETTING_FIRST = 4000;
  IDM_PLUGINSETTING_LAST = IDM_PLUGINSETTING_FIRST + 999;

  ZoomTable: array [0..15] of Double = (
    0.125, 0.25, 0.5, 0.667, 1.0, 1.25, 1.5, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0,
    12.0, 16.0
  );

  KeyScroll = 16;
  KeyMapFile = 'keymap';
  WallpaperFile = 'wallpaper.bmp';
  PluginListFile = 'pluginlist';

var
  ProfDir: string;
  MainForm: TNkVMainForm;
  Cmd: TSCmd;
  Image: TNkVImage;
  Archive: TArchive;
  FileList: TFileList;
  FileListAr: TFileList;
  Plugins: TSStrList;
  PicInfoDlg: TPicInfoDlg;
  InfoDlg: HWND;
  InfoDlgIcon: HICON;
  ViewLastFile: Boolean;

procedure LoadPluginList(const FileName: string);
var
  s: string;
begin
  if IsFile(FileName) then
    with TSConfFile.Create(FileName) do
      try
        while GetLineNoParse(s) do
          Plugins.Add(s);
      finally
        Free;
      end
  else
    with TListDir.CreateWithMask(DirName(ParamStr(0)), '*.spi') do
      try
        while Each(s) do
          if not IsDir then
            Plugins.Add(s);
      finally
        Free;
      end;
end;

{ TFileList }

constructor TFileList.Create;
begin
  FItems := TSList.Create;
end;

destructor TFileList.Destroy;
begin
  Clear;
  FItems.Free;
end;

procedure TFileList.Clear;
var
  i: Integer;
begin
  for i := 0 to FItems.Count - 1 do
    Dispose(PFileItem(FItems[i]));
  FItems.Clear;
end;

function TFileList.Get(Index: Integer): PFileItem;
begin
  Result := PFileItem(FItems[Index]);
end;

function TFileList.GetCount: Integer;
begin
  Result := FItems.Count;
end;

procedure TFileList.Add(Item: PFileItem);
begin
  FItems.Add(Item);
end;

function FileListCompare(Item1, Item2: Pointer): Integer;
begin
  Result := StrColl(PFileItem(Item1).Name, PFileItem(Item2).Name);
end;

procedure TFileList.Update(const Dir: string; const Ar: TArchive);
var
  Item: PFileItem;
  Name: string;
begin
  Clear;
  if Assigned(Ar) then
  begin
    with Ar do
      while EachName(Name) do
        if Info[Name].FileSize > 0 then
        begin
          New(Item);
          Item.Name := Name;
          Item.Loadable := True;
          Add(Item);
        end;
  end else
  begin
    with  TListDir.Create(Dir) do
      try
        while Each(Name) do
          if not IsDir then
          begin
            New(Item);
            Item.Name := Dir + '\' + Name;
            Item.Loadable := True;
            Add(Item);
          end;
      finally
        Free;
      end;
  end;
  FItems.Sort(FileListCompare);
end;

function TFileList.IndexOf(const FileName: string): Integer;
var
  i: Integer;
begin
  for i := 0 to FItems.Count - 1 do
    if Items[i].Name = FileName then
    begin
      Result := i;
      Exit;
    end;
  Result := -1;
end;


{ TPicInfoDlg }

procedure TPicInfoDlg.SetItems;
  function BytesFormat(Value: Integer): string;
  begin
    Result := UnitFormat(Value);
    case Result[Length(Result)] of
      'G': Result := Copy(Result, 1, Length(Result)-1) + ' GB';
      'M': Result := Copy(Result, 1, Length(Result)-1) + ' MB';
      'K': Result := Copy(Result, 1, Length(Result)-1) + ' KB';
    else
      Result := Result + ' bytes';
    end;
  end;

  function GetIcon(const FileName: string): HICON;
  var
    id: Word;
  begin
    // id = 1ɂbmpŃTlCiconł͂Ȃʂiconۂ
    id := 1;
    Result := ExtractAssociatedIcon(HInstance, PChar(FileName), id);
  end;

  function FileTimeToLocalTimeStr(const FileTime: TFileTime): string;
  var
    ft: TFileTime;
    st: TSystemTime;
  begin
   if (FileTime.dwLowDateTime = 0) and (FileTime.dwHighDateTime = 0) then
     Result := ''
   else begin
     FileTimeToLocalFileTime(FileTime, ft);
     FileTimeToSystemTime(ft, st);
     Result := DateStr(st) + ' ' + TimeStr(st);
   end;
  end;
var
  dir, fname: string;
begin
  with MainForm do
  begin
    if FileName = '' then
    begin
      dir := '';
      fname := '<STDIN>';
      InfoDlgIcon := 0;
    end
    else if Assigned(Archive) then
    begin
      dir := Archive.FileName;
      fname := FileName;
      InfoDlgIcon := GetIcon(Archive.FileName);
    end else
    begin
      dir := DirName(FileName);
      fname := BaseName(FileName);
      InfoDlgIcon := GetIcon(FileName);
    end;
  end;

  Items[IDC_FILENAME].Text := fname;
  Items[IDC_LOCATION].Text := dir;
  Items[IDC_SIZE].Text := IStr(Image.Width) + 'x' + IStr(Image.Height);
  Items[IDC_BPP].Text := IStr(Image.Bpp) + ' bits/pixel';
  Items[IDC_FORMAT].Text := Image.FormatName;
  Items[IDC_FILESIZE].Text := BytesFormat(Image.OrgSize)+' ('+NumFormat(Image.OrgSize)+' bytes)';
  Items[IDC_CREATED].Text := FileTimeToLocalTimeStr(MainForm.FileCTime);
  Items[IDC_MODIFIED].Text := FileTimeToLocalTimeStr(MainForm.FileMTime);
  Items[IDC_ACCESSED].Text := FileTimeToLocalTimeStr(MainForm.FileATime);
  Items[IDC_MEMORY].Text := BytesFormat(Image.BitsSize)+' ('+NumFormat(Image.BitsSize)+' bytes)';
  Items[IDC_PLUGIN].Text := Image.FilterName;
  Items[IDC_PLUGINFILE].Text := Image.FilterFilename;
  Items[IDC_TEXTINFOLABEL].Enabled := Image.Comment <> '';
  Items[IDC_TEXTINFO].Text := Image.Comment;
  InvalidateRect(Items[IDC_ICON].Handle, nil, False);
end;

procedure TPicInfoDlg.WMInitDialog(var Message: TWMInitDialog);
begin
  SetItems;
  Message.Result := 1;
end;

procedure TPicInfoDlg.WMCommand(var Message: TWMCommand);
begin
  with Message do
    case ItemID of
      IDOK, IDCANCEL:
        begin
          Result := 1;
          Close;
        end;
    end;
end;

procedure TPicInfoDlg.WMDrawItem(var Message: TWMDrawItem);
begin
  with Message do
    if Ctl = IDC_ICON then
      with DrawItemStruct^ do
      begin
        FillRect(hDC, rcItem, GetSysColorBrush(COLOR_3DFACE));
        DrawIcon(hDC, 0, 0, InfoDlgIcon);
        Result := 1;
      end;
end;


{ TNkVMainForm }

var
  FormList: TSList;
function EnumWindowsProc(hWnd: HWND; lParam: LPARAM): BOOL; stdcall;
var
  buf: array[0..255] of Char;
  AClassName: string;
begin
  GetClassName(hWnd, buf, SizeOf(buf));
  AClassName := MainForm.ClassName;
  if buf = AClassName then
    FormList.Add(Pointer(hWnd));
  Result := True;
end;

function Compare(Item1, Item2: Pointer): Integer;
begin
  if Integer(Item1) > Integer(Item2) then
    Result := 1
  else if Integer(Item1) < Integer(Item2) then
    Result := -1
  else
    Result := 0;
end;

procedure ActivateNextForm(Prev: Boolean);
var
  i, cur: Integer;
begin
  FormList := TSList.Create;
  EnumWindows(@EnumWindowsProc, 0);
  FormList.Sort(Compare);
  cur := 0;
  for i := 0 to FormList.Count-1 do
  begin
    if HWND(FormList[i]) = MainForm.Handle then
    begin
      cur := i;
      Break;
    end;
  end;
  if Prev then
  begin
    Dec(cur);
    if cur < 0 then
      cur := FormList.Count - 1;
  end else
  begin
    Inc(cur);
    if cur = FormList.Count then
      cur := 0;
  end;
  SetForegroundWindow(HWND(FormList[cur]));
  FormList.Free;
end;

function TNkVMainForm.LoadFile(const Path: string): Boolean;
begin
  FileList.Clear;
  Archive.Free;
  Archive := nil;
  if IsDir(Path) then
  begin
    FileList.Update(Path, nil);
    FileList.CurPos := -1;
    Result := LoadNext(FileList, nil, +1);
    if not Result then
      Error(Path, 'Loadable file not found in directory.');
  end else
    Result := Load(Path, nil, False);
end;

function TNkVMainForm.LoadNextFile(Direction: Integer; LeaveAr: Boolean): Boolean;
begin
  if Assigned(Archive) then
  begin
    Result := False;
    if not LeaveAr then
      Result := LoadNext(FileListAr, Archive, Direction);
    if not Result then
    begin
      if FileList.Count = 0 then
      begin
        FileList.Update(DirName(Archive.FileName), nil);
        FileList.CurPos := FileList.IndexOf(Archive.FileName);
      end;
      if (not LeaveAr) and (Direction < 0) then
        ViewLastFile := True;
      Result := LoadNext(FileList, nil, Direction);
      ViewLastFile := False;
    end;
  end else
  begin
    if FileList.Count = 0 then
    begin
      FileList.Update(DirName(FileName), nil);
      FileList.CurPos := FileList.IndexOf(FileName);
    end;
    if (not LeaveAr) and (Direction < 0) then
      ViewLastFile := True;
    Result := LoadNext(FileList, nil, Direction);
    ViewLastFile := False;
  end;
end;

function TNkVMainForm.LoadNext(const AFileList: TFileList; const Ar: TArchive;
  Direction: Integer): Boolean;
var
  i: Integer;
begin
  if Direction > 0 then
  begin
    for i := AFileList.CurPos+1 to AFileList.Count-1 do
      with AFileList[i]^ do
        if Loadable then
        begin
          Loadable := Load(Name, Ar, True);
          if Loadable then
          begin
            AFileList.CurPos := i;
            Result := True;
            Exit;
          end;
        end;
  end else
    for i := AFileList.CurPos-1 downto 0 do
      with AFileList[i]^ do
        if Loadable then
        begin
          Loadable := Load(Name, Ar, True);
          if Loadable then
          begin
            AFileList.CurPos := i;
            Result := True;
            Exit;
          end;
        end;
  Result := False;
end;

function LookupFilter(Buf: PChar; BufSize: Integer): TImgFilter;
var
  i: Integer;
begin
  Result := TDibFilter.Create;
  if Result.Test(Buf, BufSize) then
    Exit;
  Result.Free;

  for i := 0 to Plugins.Count-1 do
  begin
    Result := TSpiFilter.Create(Plugins[i]);
    if Result.Test(Buf, BufSize) then
      Exit;
    Result.Free;
  end;

  Result := nil;
end;

function TNkvMainForm.Load(const AFileName: string; const AAr: TArchive; Quiet: Boolean): Boolean;
const
  TestBufSize = 2048;
  StdinBufSize = 32768;
var
  fsize, readsize: Integer;
  filter: TImgFilter;
  buf: PChar;
  oldcur: HCURSOR;
  ar: TArchive;
  flist: TFileList;
  fd: TWin32FindData;
  isstdin: Boolean;
begin
  Result := False;
  buf := nil;
  oldcur := SetCursor(LoadCursor(0, IDC_WAIT));
  try
    if Assigned(AAr) then
    begin
      fsize := AAr[AFileName].FileSize;
      GetMem(buf, fsize+TestBufSize);
      try
        FillChar(buf^, TestBufSize, 0);
        AAr.Extract(AFileName, buf, fsize);
        // TODO: check error
        filter := LookupFilter(buf, fsize);
        try
          if not Assigned(filter) then
          begin
            if not Quiet then
              Error(AFileName, 'Unknown format.');
            Exit;
          end;
          if (filter is TSpiFilter) and (filter as TSpiFilter).IsArchive then
            Exit; 
          if not Image.ReadMem(buf, fsize, filter) then
          begin
            Error(AFileName, 'Load failed.'#13#10#13#10+'('+filter.ErrorMsg+')');
            Exit;
          end;
        finally
          filter.Free;
        end;
      finally
        FreeMem(buf);
      end;
    end else
    begin
      isstdin := AFileName = '';
      if not isstdin then
        GetFindData(AFileName, fd);
      with TSFile.Create(AFileName, 'r') do
        try
          if not Opened then
          begin
            if isstdin then
              Result := True // Win9x̓_CNgĂȂSTDINopenłȂ
            else
              Error(Name, SysError(GetLastError));
            Exit;
          end;

          GetMem(buf, TestBufSize);
          FillChar(buf^, TestBufSize, 0);
          readsize := Read(buf^, TestBufSize);
          if readsize < 0 then
          begin
            if isstdin then
              Result := True // STDINǂ߂ȂƂ̓G[ɂȂ
            else
              Error(Name, SysError(GetLastError));
            Exit;
          end;

          filter := LookupFilter(buf, readsize);
          try
            if not Assigned(filter) then
            begin
              if not Quiet then
                Error(Name, 'Unknown format.');
              Exit;
            end;

            // Archive file
            if filter is TSpiFilter then
              with filter as TSpiFilter do
                if IsArchive then
                begin
                  ar := TArchive.Create(FileName, AFileName);
                  if not ar.Opened then
                  begin
                    ar.Free;
                    if not Quiet then
                      Error(Name, 'Cannot get file list in archive.');
                    Exit;
                  end;
                  flist := TFileList.Create;
                  flist.Update('', ar);
                  if ViewLastFile then
                  begin
                    flist.CurPos := flist.Count;
                    Result := LoadNext(flist, ar, -1);
                  end else
                  begin
                    flist.CurPos := -1;
                    Result := LoadNext(flist, ar, +1);
                  end;
                  if Result then
                  begin
                    Archive.Free;
                    Archive := ar;
                    FileListAr := flist;
                    // ArchiveXVĂȂarchive fileXVȂ
                    UpdateCaption;
                  end
                  else begin
                    flist.Free;
                    ar.Free;
                    if not Quiet then
                      Error(Name, 'Loadable file not found in archive.');
                  end;
                  Exit;
                end;

            if isstdin then
            begin
              fsize := readsize;
              while True do
              begin
                ReallocMem(buf, fsize+StdinBufSize);
                readsize := Read(buf[fsize], StdinBufSize);
                if readsize <= 0 then
                  Break;
                Inc(fsize, readsize);
              end;
            end else
            begin
              fsize := Size;
              ReallocMem(buf, fsize);
              Read(buf[readsize], fsize-readsize);
            end;
            if not Image.ReadMem(buf, fsize, filter) then
            begin
              Error(Name, 'Load failed.'#13#10#13#10+'('+filter.ErrorMsg+')');
              Exit;
            end;

            Archive.Free;
            Archive := nil;
          finally
            filter.Free;
          end;
        finally
          if buf <> nil then
            FreeMem(buf);
          Free; // TSFile
        end;
    end;
  finally
    SetCursor(oldcur);
  end;

  FileName := AFileName;
  if Assigned(AAr) then
  begin
    FileCTime.dwLowDateTime := 0;
    FileCTime.dwHighDateTime := 0;
    FileMTime := AAr[FileName].DateTime;
    FileATime.dwLowDateTime := 0;
    FileATime.dwHighDateTime := 0;
  end
  else if FileName <> '' then
  begin
    FileCTime := fd.ftCreationTime;
    FileMTime := fd.ftLastWriteTime;
    FileATime := fd.ftLastAccessTime;
  end;
  ViewLeft := 0;
  ViewTop := 0;
  UpdateView;
  UpdateCaption;
  Invalidate;
  if not (Stretch or FullScreen or Maximized) then
    FitWindow;
  Update;
  if IsWindow(InfoDlg) then
    PicInfoDlg.SetItems;
  Result := True;
end;

procedure TNkVMainForm.InitParams;
begin
  inherited;
  Cursor := 0;
  ClassStyle := CS_HREDRAW or CS_VREDRAW or CS_DBLCLKS;
  ExStyle := WS_EX_ACCEPTFILES;
  Brush := GetStockObject(NULL_BRUSH);
end;

procedure TNkVMainForm.CalcViewSize(var Width: Integer; var Height: Integer);
begin
  Width := ClientRect.Right;
  if Width > DispWidth then
    Width := DispWidth;
  Height := ClientRect.Bottom;
  if Height > DispHeight then
    Height := DispHeight;
end;

procedure TNkVMainForm.UpdateCaption;
var
  fname, info: string;
  dc: HDC;
  pt: TPoint;
  c: COLORREF;
  r, g, b: Integer;
begin
  if Assigned(Archive) then
    fname := BaseName(Archive.FileName) + '/' + FileName
  else if FileName = '' then
    fname := '<STDIN>'
  else
    fname := BaseName(FileName);

  if ShowColorValue then
  begin
    GetCursorPos(pt);
    pt := ScreenToClient(pt);
    dc := GetDC(Handle);
    c := GetPixel(dc, pt.x, pt.y);
    ReleaseDC(Handle, dc);
    pt.x := Trunc((pt.x+ViewLeft) / ZoomValue);
    pt.y := Trunc((pt.y+ViewTop) / ZoomValue);
    r := c and $FF;
    g := c shr 8 and $FF;
    b := c shr 16 and $FF;
    info := ' [R'+IZStr(r, 3)+' G'+IZStr(g, 3)+' B'+IZStr(b, 3)+'] '
            + '(X'+IStr(pt.x)+' Y'+IStr(pt.y)+') ';
  end else
    info := ' ('+IStr(Image.Width)+'x'+IStr(Image.Height)+'x'+IStr(Image.Bpp)+') ';
  if Stretch then
    info := info + 'Stretch Mode'
  else
    info := info + IStr(Trunc(ZoomValue*100))+'%';
  Text := fname + ' ' + info + ' - ' + AppName;
end;

procedure TNkVMainForm.UpdateView;
var
  vw, vh, cw, ch: Integer;
begin
  if Stretch then
  begin
    cw := ClientRect.Right;
    ch := ClientRect.Bottom;
    if Image.Width * ch > Image.Height * cw then
    begin
      DispWidth := cw;
      DispHeight := Image.Height * DispWidth div Image.Width;
    end else
    begin
      DispHeight := ch;
      DispWidth := Image.Width * DispHeight div Image.Height;
    end;
    // Stretch[h甲Ƃ̂߂ViewLeft, ViewTop͍XVȂ
  end else
  begin
    DispWidth := Trunc(Image.Width * ZoomValue);
    DispHeight := Trunc(Image.Height * ZoomValue);

    CalcViewSize(vw, vh);

    if ViewLeft < 0 then
      ViewLeft := 0
    else if ViewLeft > (DispWidth-vw) then
      ViewLeft := DispWidth-vw;

    if ViewTop < 0 then
      ViewTop := 0
    else if ViewTop > (DispHeight-vh) then
      ViewTop := DispHeight-vh;
  end;
end;

procedure TNkVMainForm.FitWindow;
begin
  if not FullScreen then
  begin
    if Maximized then
      Restore;
    ClientRect := Rect(0, 0, DispWidth, DispHeight);
  end;
end;

procedure TNkVMainForm.Zoom(Value: Double);
begin
  if Value <= 0 then
    Exit;
  Stretch := False;
  ZoomValue := Value;
  UpdateView;
  UpdateCaption;
  Invalidate;
end;

procedure TNkVMainForm.Scroll(X, Y: Integer);
var
  vw, vh: Integer;
begin
  if Stretch then
    Exit;

  CalcViewSize(vw, vh);

  Inc(ViewLeft, X);
  if ViewLeft < 0 then
  begin
    Dec(X, ViewLeft);
    ViewLeft := 0;
  end
  else if ViewLeft > (DispWidth-vw) then
  begin
    Inc(X, DispWidth-vw-ViewLeft);
    ViewLeft := DispWidth-vw;
  end;

  Inc(ViewTop, Y);
  if ViewTop < 0 then
  begin
    Dec(Y, ViewTop);
    ViewTop := 0;
  end
  else if ViewTop > (DispHeight-vh) then
  begin
    Inc(Y, DispHeight-vh-ViewTop);
    ViewTop := DispHeight-vh;
  end;

  ScrollWindowEx(Handle, -X, -Y, nil, nil, 0, nil, SW_INVALIDATE);
  Update;
end;

procedure TNkVMainForm.DoCmd(const CmdName: string);
var
  i: Integer;
  fname: string;
begin
  if CmdName = cmStayOnTop then
    StayOnTop := not StayOnTop
  else if CmdName = cmHideBorder then
  begin
    if not FullScreen then
      Border := not Border;
    if Border then
      NormalCursor := LoadCursor(HInstance, 'HANDOPEN')
    else
      NormalCursor := LoadCursor(0, IDC_ARROW);
    SetCursor(NormalCursor);
  end
  else if CmdName = cmScaleFilter then
  begin
    ScaleFilter := not ScaleFilter;
    Invalidate;
  end
  else if CmdName = cmShowGrid then
  begin
    ShowGrid := not ShowGrid;
    Invalidate;
  end
  else if CmdName = cmNextWindow then
    ActivateNextForm(False)
  else if CmdName = cmPrevWindow then
    ActivateNextForm(True)
  else if CmdName = cmAboutPlugins then
    ShowAboutPlugins
  else if CmdName = cmAbout then
    ShowAbout
  else if CmdName = cmExit then
    Close;

  if not Image.Loaded then
    Exit;

  // Image Loaded
  if CmdName = cmPicInfo then
  begin
    if not (IsWindow(InfoDlg) and IsWindowVisible(InfoDlg)) then
    begin
      PicInfoDlg.ShowModeless;
      InfoDlg := PicInfoDlg.Handle;
    end;
  end
  else if CmdName = cmFitWindow then
    FitWindow
  else if CmdName = cmZoomIn then
  begin
    for i := 0 to High(ZoomTable) do
      if ZoomTable[i] > ZoomValue then
      begin
        Zoom(ZoomTable[i]);
        Break;
      end;
  end
  else if CmdName = cmZoomOut then
  begin
    for i := 0 to High(ZoomTable) do
      if ZoomTable[i] >= ZoomValue then
      begin
        if i > 0 then
          Zoom(ZoomTable[i-1]);
        Break;
      end;
  end
  else if Copy(CmdName, 1, 4) = Copy(cmZoom100, 1, 4) then
    Zoom(SInt(Copy(CmdName, 5, Length(CmdName)-5+1))/100)
  else if CmdName = cmFullScreen then
    FullScreen := not FullScreen
  else if CmdName = cmSetWallPaper then
    SetWallPaper(False)
  else if CmdName = cmSetWallPaperTile then
    SetWallPaper(True)
  else if CmdName = cmStretchMode then
  begin
    Stretch := not Stretch;
    UpdateView;
    UpdateCaption;
    Invalidate;
  end
  else if CmdName = cmSmoothZoomIn then
  begin
    for i := 0 to 4 do
    begin
      Zoom(ZoomValue+0.01);
      Update;
    end;
  end
  else if CmdName = cmSmoothZoomOut then
  begin
    for i := 0 to 4 do
    begin
      Zoom(ZoomValue-0.01);
      Update;
    end;
  end
  else if CmdName = cmScrollLeft then
    Scroll(-KeyScroll,0)
  else if CmdName = cmScrollRight then
    Scroll(KeyScroll, 0)
  else if CmdName = cmScrollUp then
    Scroll(0,-KeyScroll)
  else if CmdName = cmScrollDown then
    Scroll(0, KeyScroll)
  else if CmdName = cmScrollTop then
    Scroll(0, -DispHeight)
  else if CmdName = cmScrollBottom then
    Scroll(0, DispHeight)
  else if CmdName = cmScrollTopBottom then
  begin
    if ViewTop = DispHeight - ClientRect.Bottom then
      Scroll(0, -DispHeight)
    else
      Scroll(0, DispHeight);
  end
  else if CmdName = cmShowColorValue then
  begin
    ShowColorValue := not ShowColorValue;
    if ShowColorValue then
      NormalCursor := LoadCursor(0, IDC_CROSS)
    else if Border then
      NormalCursor := LoadCursor(HInstance, 'HANDOPEN')
    else
      NormalCursor := LoadCursor(0, IDC_ARROW);
    SetCursor(NormalCursor);
    UpdateCaption;
  end
  else if CmdName = cmSaveAs then
    SaveAs;

  if FileName = '' then
    Exit;

  // Image Loaded and not <STDIN>
  if CmdName = cmNextImage then
    LoadNextFile(+1, False)
  else if CmdName = cmPrevImage then
    LoadNextFile(-1, False)
  else if CmdName = cmNextFile then
    LoadNextFile(+1, True)
  else if CmdName = cmPrevFile then
    LoadNextFile(-1, True)
  else if CmdName = cmRefreshFileList then
    FileList.Clear
  else if CmdName = cmCopyFileName then
    CopyFileNameToClipBoard
  else if CmdName = cmReloadImage then
  begin
    if Assigned(Archive) then
    begin
      fname := Archive.FileName; // LoadFileArchivefreêŃRs[
      LoadFile(fname)
    end else
      LoadFile(FileName)
  end;
  {
  IDM_ARCHIVEFILELIST:
    begin
      for i := 0 to ArcFileList.ItemCount - 1 do
        s := s + PChar(ArcFileList[i]) + #13#10;
      MessageBox(Handle, PChar(s), AppName + '', 0);
    end;
  }
end;

procedure TNkVMainForm.ShowAbout;
begin
  AboutBox(Handle, 'About '+AppName,
    AppName+' '+Version+#13#10+
    #13#10+
    Copyright+#13#10+
    #13#10+
    License+#13#10+
    #13#10+
    'Profile Folder: '+ProfDir);
end;

procedure TNkVMainForm.PopupMenu(const Pt: TPoint);
var
  i: Integer;
  Menu, SubMenu: HMENU;
  CmdMenu: TSCmdMenu;
begin
  CmdMenu := TSCmdMenu.Create(Cmd, MenuTable);
  if not Image.Loaded then
    CmdMenu.DisableItems([
      cmPrevImage,
      cmNextImage,
      cmPrevFile,
      cmNextFile,
      cmPicInfo,
      cmStretchMode,
      cmFullScreen,
      cmFitWindow,
      cmZoom100,
      cmZoomIn,
      cmZoomOut,
       '{ZoomMenu',
      cmShowColorValue,
      cmCopyFileName,
      cmReloadImage,
      '{WallPaperMenu',
      cmSaveAs
    ]);
  if Image.Loaded and (FileName = '') then
    CmdMenu.DisableItems([
      cmPrevImage,
      cmNextImage,
      cmPrevFile,
      cmNextFile,
      cmCopyFileName,
      cmReloadImage
    ]);
  if FullScreen then
    CmdMenu.DisableItems([cmFitWindow]);
  if Plugins.Count = 0 then
    CmdMenu.DisableItems(['{PluginMenu']);

  CmdMenu.CheckItem(cmFullScreen, FullScreen);
  CmdMenu.CheckItem(cmStayOnTop, StayOnTop);
  CmdMenu.CheckItem(cmStretchMode, Stretch);
  CmdMenu.CheckItem(cmHideBorder, not Border);
  CmdMenu.CheckItem(cmShowColorValue, ShowColorValue);
  CmdMenu.CheckItem(cmScaleFilter, ScaleFilter);

  Menu := CmdMenu.MakeMenuHandle;
  CmdMenu.Free;

  SubMenu := GetSubMenu(Menu, GetMenuItemCount(Menu) - 6);
  for i := 0 to Plugins.Count-1 do
    AppendMenu(SubMenu, MF_STRING, IDM_PLUGINSETTING_FIRST+i, PChar(Plugins[i]));

  TrackPopupMenu(Menu, TPM_LEFTALIGN or TPM_LEFTBUTTON or TPM_RIGHTBUTTON,
    pt.x, pt.y, 0, Handle, nil);

  DestroyMenu(Menu);
end;

procedure TNkVMainForm.Error(const AFileName: string; const Msg: string);
begin
  MessageBox(Handle, PChar(AFileName + #13#10#13#10 + Msg), AppName,
    MB_ICONEXCLAMATION);
end;

procedure TNkVMainForm.SaveAs;
var
  fname, ext: string;
begin
  SplitExt(BaseName(FileName), fname, ext);
  if fname = '' then
    fname := 'stdin';
  with TSSaveDlg.Create(Handle, ['All Files', '*.*'], fname + '.bmp') do
    try
      if ShowModal(fname) then
      begin
        if not Image.Write(fname, DispWidth, DispHeight, ScaleFilter) then
          Error(fname, SysError(GetLastError));
      end;
    finally
      Free;
    end;
end;

procedure TNkVMainForm.SetWallPaper(Tile: Boolean);
var
  fname: string;
  key: HKEY;
  data: array[0..MAX_PATH] of Char;
begin
  fname := JoinPath(ProfDir, WallpaperFile);
  if not Image.Write(fname, Image.Width, Image.Height, False) then
  begin
    Error(fname, SysError(GetLastError));
    Exit;
  end;

  if Tile then
    data := '1'
  else
    data := '0';
  if RegOpenKeyEx(HKEY_CURRENT_USER, 'Control Panel\desktop', 0, KEY_WRITE, key) = ERROR_SUCCESS then
  begin
    RegSetValueEx(key, 'TileWallPaper', 0, REG_SZ, @data, 1);
    RegCloseKey(key);
  end;
  SystemParametersInfo(SPI_SETDESKWALLPAPER, 0, PChar(fname),
    SPIF_UPDATEINIFILE or SPIF_SENDWININICHANGE);
end;

procedure TNkVMainForm.CopyFileNameToClipBoard;
var
  s: string;
  h: HGLOBAL;
  p: PChar;
begin
  if Assigned(Archive) then
    s := Archive.FileName
  else
    s := FileName;
  h := GlobalAlloc(GHND, Length(s)+1);
  p := GlobalLock(h);
  Move(s[1], p^, Length(s));
  GlobalUnlock(h);
  if OpenClipBoard(Handle) then
  begin
    EmptyClipBoard;
    SetClipBoardData(CF_TEXT, h);
    CloseClipBoard;
  end else
    GlobalFree(h);
end;

procedure TNkVMainForm.ShowPluginConfig(const PluginFileName: string);
begin
  with TSpiFilter.Create(PluginFileName) do
    try
      if Handle = 0 then
      begin
        Error(PluginFileName, 'Load failed.');
        Exit;
      end;
      if not ShowConfigDlg(Self.Handle) then
        Error(PluginFileName, ErrorMsg);
    finally
      Free;
    end;
end;

procedure TNkVMainForm.ShowAboutPlugins;
var
  i: Integer;
  s: string;
begin
  s := IStr(Plugins.Count) + ' plugin(s)'#13#10#13#10;
  for i := 0 to Plugins.Count-1 do
    with TSpiFilter.Create(Plugins[i]) do
      if Handle = 0 then
        s := s + '(not found)'#13#10
      else
        s := s + FileName + ': ' + Name + #13#10;
  MsgBox(Handle, AppName + ' - About Plugins', s);
end;

procedure TNkVMainForm.WMPaint(var Message: TWMPaint);
var
  dc: HDC;
  ps: TPaintStruct;
  vl, vt, vw, vh, cw, ch, l, t, x, y, z: Integer;
  br, kbr: HBRUSH;
begin
  dc := BeginPaint(Handle, ps);

  br := GetStockObject(GRAY_BRUSH);

  if not Image.Loaded then
  begin
    FillRect(dc, ClientRect, br);
    Exit;
  end;

  cw := ClientRect.Right;
  ch := ClientRect.Bottom;
  CalcViewSize(vw, vh);

  if Stretch then
  begin
    vl := 0;
    vt := 0;
  end else
  begin
    vl := ViewLeft;
    vt := ViewTop;
  end;

  l := 0;
  t := 0;
  if FullScreen then
  begin
    if vw < cw then
      l := (cw - vw) div 2;
    if vh < ch then
      t := (ch - vh) div 2;
  end;

  Image.Draw(dc, l, t, vw, vh, vl, vt, DispWidth, DispHeight, ScaleFilter);

  if ShowGrid and (ZoomValue >= 4) then
  begin
    kbr := GetStockObject(BLACK_BRUSH);
    z := Trunc(ZoomValue);
    y := t + (vt+z-1) div z * z - vt;
    while y < (t+vh) do
    begin
      FrameRect(dc, Rect(l-1, y, l+vw+1, y+z+1), kbr);
      Inc(y, z*2);
    end;
    x := l + (vl+z-1) div z * z - vl;
    while x < (l+vw) do
    begin
      FrameRect(dc, Rect(x, t-1, x+z+1, t+vh+1), kbr);
      Inc(x, z*2);
    end;
  end;

  if vw < cw then
  begin
    FillRect(dc, Rect(l+vw, t, cw, t+vh), br); // right
    FillRect(dc, Rect(0, t, l, t+vh), br); // left
  end;
  if vh < ch then
  begin
    FillRect(dc, Rect(0, t+vh, cw, ch), br); // bottom
    FillRect(dc, Rect(0, 0, cw, t), br); // top
  end;

  EndPaint(Handle, ps);
end;

procedure TNkVMainForm.WMSetCursor(var Message: TWMSetCursor);
begin
  with Message do
    if HitTest = HTCLIENT then
    begin
      SetCursor(NormalCursor);
      Result := 1;
    end;
  inherited;
end;

procedure TNkVMainForm.WMLButtonDown(var Message: TWMLButtonDown);
begin
  if Border or FullScreen or (not Border and (GetKeyState(VK_SHIFT) < 0)) then
  begin
    SetCapture(Handle);
    SetCursor(GrabCursor);
    GetCursorPos(OldPt);
  end else
  begin
    ReleaseCapture;
    Perform(WM_NCLBUTTONDOWN, HTCAPTION, 0);
  end;
end;

procedure TNkVMainForm.WMLButtonUp(var Message: TWMLButtonUp);
begin
  ReleaseCapture;
  SetCursor(NormalCursor);
end;

procedure TNkVMainForm.WMMouseMove(var Message: TWMMouseMove);
var
  pt: TPoint;
begin
  if Message.Keys and MK_LBUTTON <> 0 then
  begin
    // MessageɊ܂܂}EXJ[\W̓}EXJ[\windowÔƂ
    // lł͂Ȃ̂GetCursorPosŎ擾
    GetCursorPos(pt);
    Scroll(OldPt.x - pt.x, OldPt.y - pt.y);
    OldPt := pt;
  end
  else if ShowColorValue then
    UpdateCaption;
end;

procedure TNkVMainForm.WMMouseWheel(var Message: TMessage);
begin
  with Message do
    if GetKeyState(VK_LBUTTON) < 0 then
    begin
      if WParam > 0 then
        DoCmd(cmSmoothZoomIn)
      else
        DoCmd(cmSmoothZoomOut);
    end else
    begin
      if WParam > 0 then
        DoCmd(cmPrevImage)
      else
        DoCmd(cmNextImage);
    end;
end;

procedure TNkVMainForm.WMLButtonDblClk(var Message: TWMLButtonDblClk);
begin
  DoCmd(cmFitWindow);
end;

procedure TNkVMainForm.WMSize(var Message: TWMSize);
begin
  UpdateView;
end;

procedure TNkVMainForm.WMCommand(var Message: TWMCommand);
begin
  with Message do
    case ItemID of
      IDM_PLUGINSETTING_FIRST .. IDM_PLUGINSETTING_LAST:
        ShowPluginConfig(Plugins[ItemID - IDM_PLUGINSETTING_FIRST]);
    else
      DoCmd(Cmd.Name(ItemID));
    end;
end;

procedure TNkVMainForm.WMContextMenu(var Message: TMessage);
var
  pt: TPoint;
begin
  if GetContextMenuPos(Handle, Message.LParam, 0, 0, pt) then
    PopupMenu(pt)
  else
    inherited;
end;

procedure TNkVMainForm.WMDropFiles(var Message: TWMDropFiles);
var
  fname: string;
begin
  with Message do
  begin
    SetString(fname, nil, DragQueryFile(Drop, 0, nil, 0));
    DragQueryFile(Drop, 0, PChar(fname), Length(fname)+1); // +1: null
    DragFinish(Drop);
  end;
  LoadFile(fname);
end;

begin
  ProfDir := SProf.ProfDir('NKTOOLSPROF', 'nkv');
  Cmd := TSCmd.Create(CmdList, DefaultKeys, JoinPath(ProfDir, KeyMapFile));
  Image := TNkVImage.Create;
  FileList := TFileList.Create;
  Plugins := TSStrList.Create;
  MainForm := TNkVMainForm.CreateDefaultPos(AppName, 320, 240);
  try
    if MainForm.Handle = 0 then
      Halt(255); // TODO: show error msg

    SetMainForm(MainForm);

    LoadPluginList(JoinPath(ProfDir, PluginListFile));

    MainForm.NormalCursor := LoadCursor(HInstance, 'HANDOPEN');
    MainForm.GrabCursor := LoadCursor(HInstance, 'HANDHOLD');
    MainForm.ZoomValue := 1.0;

    if not MainForm.LoadFile(LongFileName(ParamStr(1))) then
      Exit;

    PicInfoDlg := TPicInfoDlg.CreateByResource(MainForm.Handle, 'INFODIALOG');

    MainForm.Show;

    ExitCode := MessageLoopWithAccelAndDialog(MainForm.Handle, Cmd.AccHandle, InfoDlg);
  finally
    PicInfoDlg.Free;
    Plugins.Free;
    FileList.Free;
    Archive.Free;
    Image.Free;
    MainForm.Free;
    Cmd.Free;
  end;
end.


